# -*- coding: utf-8 -*-
"""
Benchmarks for primitive/low-level array operations.
"""
__all__ = [
    "bench_astype",
    "bench_astype_numba",
    "bench_astype_numpy",
    "bench_bool_index",
    "bench_bool_index_numpy",
    "bench_compare_ops",
    "bench_compare_ops_numpy",
    "bench_mbget",
    "bench_mbget_numba",
    # comparisons
    "compare_astype",
    "compare_bool_index",
    "compare_compare_ops",
    "compare_mbget",
]

import itertools
import logging
import operator
from typing import List

import numba as nb
import numpy as np
from numpy.random import default_rng

from ..config import get_global_settings
from ..rt_dataset import Dataset
from ..rt_enum import NumpyCharTypes, TypeRegister
from ..rt_numpy import empty
from ..rt_utils import _mbget_2dims, mbget  # , mbget_numba
from .benchmark import _timestamp_funcs
from .rand_data import rand_array, rand_fancyindex
from .runner import create_comparison_dataset, create_trial_dataset

logger = logging.getLogger(__name__)
"""The logger for this module."""

timestamper = _timestamp_funcs["get_nano_time"]
"""The timestamping function to use in benchmarks."""

# TODO: Additional benchmarks which would be useful for riptable development and comparison to other frameworks:
#   * mbget vs. numpy fancy indexing (only on valid array indices -- -len(arr) <= x < len(arr))
#   * mbget vs. numba-based equivalent to look for compiled code optimizations + thread scaling
#   * indexing with a boolean mask (riptable vs. numba)
#   * array conversion (i.e. elementwise type conversion) (arr1.astype(np.float32))
#       * make sure to include the self-conversion case so that we look for optimizations there (like just calling memcpy)
#   * equality and comparisons
#       * elementwise array equality (arr1 == arr2)
#       * array vs. scalar equality (arr == 123, arr == "foo", arr != '', etc.)
#       * elementwise array comparison (arr1 < arr2)
#       * array vs. scalar comparison (arr1 < 1.23, arr > 123, etc.)
#       * it would also be useful (for demonstration purposes) to demo here how much faster these operations
#         are on a string categorical compared to a normal array of strings (like the Categorical's .expand_array).
#   * conversion-assignment, e.g. result[:] = arr[:]


def mbget_numba(aValues, aIndex) -> np.ndarray:
    """
    Re-implementation of the 'mbget' fancy-indexing function with numba, for comparison with the riptide_cpp implementation.

    Parameters
    ----------
    aValues
    aIndex

    Returns
    -------

    """
    # make sure a aValues and aIndex are both numpy arrays
    if isinstance(aValues, (list, tuple)):
        aValues = TypeRegister.FastArray(aValues)

    if isinstance(aIndex, (list, tuple)):
        aIndex = TypeRegister.FastArray(aIndex)

    if not isinstance(aValues, np.ndarray) or not isinstance(aIndex, np.ndarray):
        raise TypeError(f"Values and index must be numpy arrays. Got {type(aValues)} {type(aIndex)}")

    elif aValues.dtype.char == "O":
        raise TypeError(f"mbget does not support object types")

    elif aIndex.dtype.char not in NumpyCharTypes.AllInteger:
        raise TypeError(f"indices provided to mbget must be an integer type not {aIndex.dtype}")

    if aValues.ndim == 2:
        return _mbget_2dims(aValues, aIndex)

    # TODO: probably need special code or parameter to set custom default value for NAN_TIME
    if aValues.dtype.char in NumpyCharTypes.AllInteger + NumpyCharTypes.AllFloat:
        result = _mbget_numeric(aValues, aIndex)
    elif aValues.dtype.char in "SU":
        result = _mbget_string(aValues, aIndex)
    else:
        raise Exception(f"mbget can't operate on an array of this type: {aValues.dtype}")

    result = TypeRegister.newclassfrominstance(result, aValues)

    return result


def _mbget_numeric(aValues, aIndex) -> np.ndarray:
    result = empty(len(aIndex), dtype=aValues.dtype)

    # Choose different implementation for signed vs. unsigned dtype.
    # See comment in mbget_string for details.
    _mbget_numeric_impl = _mbget_numeric_unsigned_impl if aIndex.dtype.kind == "u" else _mbget_numeric_signed_impl
    _mbget_numeric_impl(aValues, aIndex, result, result.inv)
    return result


numba_parallel = nb.njit(cache=get_global_settings().enable_numba_cache, parallel=True, nogil=True)


@numba_parallel
def _mbget_numeric_signed_impl(aValues, aIndex, result, default_val):
    num_elmnts = len(aValues)
    for i in nb.prange(aIndex.shape[0]):
        # This has one less branch (in the code) than the riptide_cpp implementation of mbget,
        # because numba handles the negative/wraparound indexing for us. So the conditional logic
        # to handle the negative indexing is still there; it may or may not be in the generated
        # machine code depending on how numba chooses to generate it.
        index = aIndex[i]
        result[i] = aValues[index] if -num_elmnts <= index < num_elmnts else default_val


@numba_parallel
def _mbget_numeric_unsigned_impl(aValues, aIndex, result, default_val):
    num_elmnts = len(aValues)
    for i in nb.prange(aIndex.shape[0]):
        # This has one less branch (in the code) than the riptide_cpp implementation of mbget,
        # because numba handles the negative/wraparound indexing for us. So the conditional logic
        # to handle the negative indexing is still there; it may or may not be in the generated
        # machine code depending on how numba chooses to generate it.
        index = aIndex[i]
        result[i] = aValues[index] if index < num_elmnts else default_val


# not using a default value here since we're handling byte strings only (default val. is 0)
def _mbget_string(aValues, aIndex) -> np.ndarray:
    result = empty(len(aIndex), dtype=aValues.dtype)
    itemsize = aValues.dtype.itemsize // 1  # ASCII

    # Choose different implementation for signed vs. unsigned dtype.
    # This is both for performance reasons and also because if we try to use the signed implementation
    # with unsigned integer types, numba ends up doing extra/unnecessary conversions so the performance
    # is poor; for that same reason, numba fails with an error on the uint64 type since it tries to cast
    # the index value to a float before we use it as an array index (which isn't allowed).
    # TODO: This decision could probably be pushed into the numba JIT-specialized generic so we don't need to choose here?
    _mbget_string_impl = _mbget_string_unsigned_impl if aIndex.dtype.kind == "u" else _mbget_string_signed_impl
    _mbget_string_impl(aValues.view(np.uint8), aIndex, result.view(np.uint8), itemsize)
    return result


@numba_parallel
def _mbget_string_signed_impl(aValues, aIndex, result, itemsize):  # byte array
    numstrings = aValues.shape[0] // itemsize
    for i in nb.prange(aIndex.shape[0]):
        index = aIndex[i]
        if -numstrings <= index < numstrings:
            str_idx = index if index >= 0 else numstrings + aIndex[i]
            for j in range(itemsize):
                result[itemsize * i + j] = aValues[itemsize * str_idx + j]

        else:
            for j in range(itemsize):
                result[itemsize * i + j] = 0


@numba_parallel
def _mbget_string_unsigned_impl(aValues, aIndex, result, itemsize):  # byte array
    numstrings = aValues.shape[0] // itemsize
    for i in nb.prange(aIndex.shape[0]):
        index = aIndex[i]
        if index < numstrings:
            for j in range(itemsize):
                result[itemsize * i + j] = aValues[itemsize * index + j]

        else:
            for j in range(itemsize):
                result[itemsize * i + j] = 0


def astype_numba(arr, dst_dtype):
    # only supports numeric-to-numeric type conversions
    if arr.dtype.char in "SU" or dst_dtype.char in "SU":
        raise Exception(f"Only numeric-to-numeric type conversions are supported.")
    result = empty(arr.shape[0], dtype=dst_dtype)
    _astype_numba(arr, result)
    return result


# numba seems to emit poor quality code for this simple loop, and the performance is
# massively worse when parallel=True is specified. (Tested with numba 0.48, 0.50.1)
# Manually splitting the loop so the input data is chunked does not improve the performance either.
@numba_parallel
def _astype_numba(arr, result):
    for i in nb.prange(len(arr)):
        # conversion occurs implicitly, and numba only supports conversion
        # between arrays of numeric types.
        result[i] = arr[i]


def bench_bool_index(**kwargs) -> Dataset:
    warmup_iters = 0
    iters = 21
    rng_seeds = [12345]
    data_dtypes = [
        np.int16,
        np.int32,
        np.float64,
        # TODO: Enable these additional data types; they're somewhat slow though, so we'd only want to
        #       run them under a 'detailed' / 'long-running' scenario
        # np.dtype('S4'),
        # np.dtype('S10'),
        # np.dtype('<U8')
    ]
    data_lengths = [
        100,
        1_000,
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        # TODO: Add 100M, 1G and 2G -- these need to be optional since smaller machines will run out of memory
        #       and also take longer than typical trials
    ]
    true_ratios = [0.0, 0.2, 0.4, 0.6, 0.8, 1.0]

    setup_params = itertools.product(rng_seeds, data_dtypes, data_lengths, true_ratios)

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for (
        rng_seed,
        data_dtype,
        data_length,
        true_ratio,
    ) in setup_params:
        # HACK: Until we have a better approach for supporting non-rectangular parameter spaces,
        #       or otherwise being able to skip certain combinations of parameters (e.g. because
        #       they're invalid, non-realistic, or otherwise don't make sense).
        # if np.iinfo(index_dtype).max < data_length:
        #     continue

        #
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        data_array = rand_array(rng, data_length, dtype=np.dtype(data_dtype))
        mask = rng.random(data_length) < true_ratio

        # Sweep over other parameters that aren't required by the setup phase.
        other_params = [None]
        for _ in other_params:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    _ = data_array[mask]

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "data_dtype": np.dtype(data_dtype),
                    "data_length": data_length,
                    "true_ratio": true_ratio,
                    # Other parameters
                    # (None)
                },
            )
            benchmark_data.append(trial_data)

        # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_bool_index_numpy(**kwargs) -> Dataset:
    warmup_iters = 0
    iters = 21
    rng_seeds = [12345]
    data_dtypes = [
        np.int16,
        np.int32,
        np.float64,
        # TODO: Enable these additional data types; they're somewhat slow though, so we'd only want to
        #       run them under a 'detailed' / 'long-running' scenario
        # np.dtype('S4'),
        # np.dtype('S10'),
        # np.dtype('<U8')
    ]
    data_lengths = [
        100,
        1_000,
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        # TODO: Add 100M, 1G and 2G -- these need to be optional since smaller machines will run out of memory
        #       and also take longer than typical trials
    ]
    true_ratios = [0.0, 0.2, 0.4, 0.6, 0.8, 1.0]

    setup_params = itertools.product(rng_seeds, data_dtypes, data_lengths, true_ratios)

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for (
        rng_seed,
        data_dtype,
        data_length,
        true_ratio,
    ) in setup_params:
        # HACK: Until we have a better approach for supporting non-rectangular parameter spaces,
        #       or otherwise being able to skip certain combinations of parameters (e.g. because
        #       they're invalid, non-realistic, or otherwise don't make sense).
        # if np.iinfo(index_dtype).max < data_length:
        #     continue

        #
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        data_array = rand_array(rng, data_length, dtype=np.dtype(data_dtype))
        if hasattr(data_array, "_np"):
            data_array = data_array._np
        mask = rng.random(data_length) < true_ratio

        # Sweep over other parameters that aren't required by the setup phase.
        other_params = [None]
        for _ in other_params:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    _ = data_array[mask]

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "data_dtype": np.dtype(data_dtype),
                    "data_length": data_length,
                    "true_ratio": true_ratio,
                    # Other parameters
                    # (None)
                },
            )
            benchmark_data.append(trial_data)

        # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_mbget(**kwargs) -> Dataset:
    # TODO: Add additional dimensions:
    #  * number of threads
    #  * recycler on/off
    #  * different key multiplicity distributions (in the rand_fancyindex function)
    #  * different amounts of 'sortedness' of the fancy index (from the rand_fancyindex function)

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 0
    iters = 21  # This duration of this function is (usually) fairly short, so the performance is prone to random noise -- using more iterations helps

    # Parameters we'll sweep over for the benchmark.
    rng_seeds = [12345]
    data_dtypes = [
        np.int16,
        np.int32,
        np.float64,
        # TODO: Enable these additional data types; they're somewhat slow though, so we'd only want to
        #       run them under a 'detailed' / 'long-running' scenario
        np.dtype("S11"),
    ]
    index_dtypes = [
        np.int8,
        np.uint8,
        np.int16,
        np.uint16,
        np.int32,
        np.uint32,
        np.int64,
        np.uint64,
        # TODO: Add float32 / float64 once rand_fancyindex() supports them
    ]
    data_lengths = [
        100,
        1_000,
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        # TODO: Add 100M, 1G and 2G -- these need to be optional since smaller machines will run out of memory
        #       and also take longer than typical trials
    ]
    index_lengths = [10, 100, 1_000, 10_000, 100_000, 1_000_000, 10_000_000]
    invalid_ratios = [
        0.0,
        0.01,
        0.1,
        # TODO: Enable these additional values for the 'detailed' scenario
        # 0.5,
        # 0.9,
    ]

    setup_params = itertools.product(
        rng_seeds,
        data_dtypes,
        index_dtypes,
        data_lengths,
        index_lengths,
        invalid_ratios,
    )

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for (
        rng_seed,
        data_dtype,
        index_dtype,
        data_length,
        index_length,
        invalid_ratio,
    ) in setup_params:
        # HACK: Until we have a better approach for supporting non-rectangular parameter spaces,
        #       or otherwise being able to skip certain combinations of parameters (e.g. because
        #       they're invalid, non-realistic, or otherwise don't make sense).
        if np.iinfo(index_dtype).max < data_length:
            continue

        #
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        data_array = rand_array(rng, data_length, dtype=np.dtype(data_dtype), invalid_ratio=invalid_ratio)
        fancyindex = rand_fancyindex(
            rng,
            index_length,
            dtype=np.dtype(index_dtype),
            source_arr_len=data_length,
            invalid_ratio=invalid_ratio,
        )

        # Sweep over other parameters that aren't required by the setup phase.
        other_params = [None]
        for _ in other_params:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    mbget(data_array, fancyindex)

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "data_dtype": np.dtype(data_dtype),
                    "index_dtype": np.dtype(index_dtype),
                    "data_length": data_length,
                    "index_length": index_length,
                    "invalid_ratio": invalid_ratio,
                    # Other parameters
                    # (None)
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_mbget_numba(**kwargs) -> Dataset:
    # TODO: Add additional dimensions:
    #  * number of threads
    #  * recycler on/off
    #  * different key multiplicity distributions (in the rand_fancyindex function)
    #  * different amounts of 'sortedness' of the fancy index (from the rand_fancyindex function)

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 1
    iters = 21  # This duration of this function is (usually) fairly short, so the performance is prone to random noise -- using more iterations helps

    # Parameters we'll sweep over for the benchmark.
    rng_seeds = [12345]
    data_dtypes = [
        np.int16,
        np.int32,
        np.float64,
        # TODO: Enable these additional data types; they're somewhat slow though, so we'd only want to
        #       run them under a 'detailed' / 'long-running' scenario
        np.dtype("S11"),
    ]
    index_dtypes = [
        np.int8,
        np.uint8,
        np.int16,
        np.uint16,
        np.int32,
        np.uint32,
        np.int64,
        np.uint64,
        # TODO: Add float32 / float64 once rand_fancyindex() supports them
    ]
    data_lengths = [
        100,
        1_000,
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        # TODO: Add 100M, 1G and 2G -- these need to be optional since smaller machines will run out of memory
        #       and also take longer than typical trials
    ]
    index_lengths = [10, 100, 1_000, 10_000, 100_000, 1_000_000, 10_000_000]
    invalid_ratios = [
        0.0,
        0.01,
        0.1,
        # TODO: Enable these additional values for the 'detailed' scenario
        # 0.5,
        # 0.9,
    ]

    setup_params = itertools.product(
        rng_seeds,
        data_dtypes,
        index_dtypes,
        data_lengths,
        index_lengths,
        invalid_ratios,
    )

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []
    for (
        rng_seed,
        data_dtype,
        index_dtype,
        data_length,
        index_length,
        invalid_ratio,
    ) in setup_params:
        # HACK: Until we have a better approach for supporting non-rectangular parameter spaces,
        #       or otherwise being able to skip certain combinations of parameters (e.g. because
        #       they're invalid, non-realistic, or otherwise don't make sense).
        if np.iinfo(index_dtype).max < data_length:
            continue

        #
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        data_array = rand_array(rng, data_length, dtype=np.dtype(data_dtype), invalid_ratio=invalid_ratio)
        fancyindex = rand_fancyindex(
            rng,
            index_length,
            dtype=np.dtype(index_dtype),
            source_arr_len=data_length,
            invalid_ratio=invalid_ratio,
        )

        # Sweep over other parameters that aren't required by the setup phase.
        other_params = [None]
        for _ in other_params:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    mbget_numba(data_array, fancyindex)

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "data_dtype": np.dtype(data_dtype),
                    "index_dtype": np.dtype(index_dtype),
                    "data_length": data_length,
                    "index_length": index_length,
                    "invalid_ratio": invalid_ratio,
                    # Other parameters
                    # (None)
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_astype(**kwargs):
    # TODO: Add additional dimensions:
    #  * number of threads
    #  * recycler on/off
    #  * different key multiplicity distributions (in the rand_fancyindex function)
    #  * different amounts of 'sortedness' of the fancy index (from the rand_fancyindex function)

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 0
    iters = 21  # This duration of this function is (usually) fairly short, so the performance is prone to random noise -- using more iterations helps

    # Parameters we'll sweep over for the benchmark.
    rng_seeds = [12345]
    src_dtypes = [
        np.int16,
        np.int32,
        np.float64,
        # np.dtype('S11'),
    ]
    dst_dtypes = [
        np.int16,
        np.int32,
        np.float64,
        # np.dtype('S11'),
    ]
    data_lengths = [
        100,
        1_000,
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        100_000_000
        # TODO: Add 100M, 1G and 2G -- these need to be optional since smaller machines will run out of memory
        #       and also take longer than typical trials
    ]

    invalid_ratios = [
        0.0,
        0.01,
        0.1,
        # TODO: Enable these additional values for the 'detailed' scenario
        # 0.5,
        # 0.9,
    ]
    setup_params = itertools.product(
        rng_seeds,
        src_dtypes,
        dst_dtypes,
        data_lengths,
        invalid_ratios,
    )

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for (
        rng_seed,
        src_dtype,
        dst_dtype,
        data_length,
        invalid_ratio,
    ) in setup_params:
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        data_array = rand_array(rng, data_length, dtype=np.dtype(src_dtype), invalid_ratio=invalid_ratio)

        # Sweep over other parameters that aren't required by the setup phase.
        other_params = [None]
        for _ in other_params:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    data_array.astype(dtype=dst_dtype)

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "src_dtype": np.dtype(src_dtype),
                    "dst_dtype": np.dtype(dst_dtype),
                    "data_length": data_length,
                    "invalid_ratio": invalid_ratio,
                    # Other parameters
                    # (None)
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_astype_numpy(**kwargs):
    # TODO: Add additional dimensions:
    #  * number of threads
    #  * recycler on/off
    #  * different key multiplicity distributions (in the rand_fancyindex function)
    #  * different amounts of 'sortedness' of the fancy index (from the rand_fancyindex function)

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 0
    iters = 21  # This duration of this function is (usually) fairly short, so the performance is prone to random noise -- using more iterations helps

    # Parameters we'll sweep over for the benchmark.
    rng_seeds = [12345]
    src_dtypes = [
        np.int16,
        np.int32,
        np.float64,
        # np.dtype('S11'),
    ]
    dst_dtypes = [
        np.int16,
        np.int32,
        np.float64,
        # np.dtype('S11'),
    ]
    data_lengths = [
        100,
        1_000,
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        100_000_000
        # TODO: Add 100M, 1G and 2G -- these need to be optional since smaller machines will run out of memory
        #       and also take longer than typical trials
    ]

    invalid_ratios = [
        0.0,
        0.01,
        0.1,
        # TODO: Enable these additional values for the 'detailed' scenario
        # 0.5,
        # 0.9,
    ]
    setup_params = itertools.product(
        rng_seeds,
        src_dtypes,
        dst_dtypes,
        data_lengths,
        invalid_ratios,
    )

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for (
        rng_seed,
        src_dtype,
        dst_dtype,
        data_length,
        invalid_ratio,
    ) in setup_params:
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        data_array = rand_array(rng, data_length, dtype=np.dtype(src_dtype), invalid_ratio=invalid_ratio)
        if hasattr(data_array, "_np"):
            data_array = data_array._np

        # Sweep over other parameters that aren't required by the setup phase.
        other_params = [None]
        for _ in other_params:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    data_array.astype(dtype=dst_dtype)

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "src_dtype": np.dtype(src_dtype),
                    "dst_dtype": np.dtype(dst_dtype),
                    "data_length": data_length,
                    "invalid_ratio": invalid_ratio,
                    # Other parameters
                    # (None)
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_astype_numba(**kwargs):
    # TODO: Add additional dimensions:
    #  * number of threads
    #  * recycler on/off
    #  * different key multiplicity distributions (in the rand_fancyindex function)
    #  * different amounts of 'sortedness' of the fancy index (from the rand_fancyindex function)

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 1
    iters = 21  # This duration of this function is (usually) fairly short, so the performance is prone to random noise -- using more iterations helps

    # Parameters we'll sweep over for the benchmark.
    rng_seeds = [12345]
    src_dtypes = [
        np.int16,
        np.int32,
        np.float64,
    ]
    dst_dtypes = [
        np.int16,
        np.int32,
        np.float64,
    ]
    data_lengths = [
        100,
        1_000,
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        100_000_000
        # TODO: Add 100M, 1G and 2G -- these need to be optional since smaller machines will run out of memory
        #       and also take longer than typical trials
    ]
    invalid_ratios = [
        0.0,
        0.01,
        0.1,
        # TODO: Enable these additional values for the 'detailed' scenario
        # 0.5,
        # 0.9,
    ]
    setup_params = itertools.product(
        rng_seeds,
        src_dtypes,
        dst_dtypes,
        data_lengths,
        invalid_ratios,
    )

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for (
        rng_seed,
        src_dtype,
        dst_dtype,
        data_length,
        invalid_ratio,
    ) in setup_params:
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        data_array = rand_array(rng, data_length, dtype=np.dtype(src_dtype), invalid_ratio=invalid_ratio)

        # Sweep over other parameters that aren't required by the setup phase.
        other_params = [None]
        for _ in other_params:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    astype_numba(data_array, np.dtype(dst_dtype))

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "src_dtype": np.dtype(src_dtype),
                    "dst_dtype": np.dtype(dst_dtype),
                    "data_length": data_length,
                    "invalid_ratio": invalid_ratio
                    # Other parameters
                    # (None)
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_compare_ops(**kwargs):
    # TODO: Add additional dimensions:
    #  * number of threads
    #  * recycler on/off
    #  * different key multiplicity distributions (in the rand_fancyindex function)
    #  * different amounts of 'sortedness' of the fancy index (from the rand_fancyindex function)

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 0
    iters = 21  # This duration of this function is (usually) fairly short, so the performance is prone to random noise -- using more iterations helps

    # Parameters we'll sweep over for the benchmark.
    rng_seeds = [12345]
    arr1_dtypes = [
        np.int16,
        np.int32,
        np.float64,
    ]
    arr2_dtypes = [
        np.int16,
        np.int32,
        np.float64,
    ]
    data_lengths = [
        100,
        1_000,
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        # TODO: Add 100M, 1G and 2G -- these need to be optional since smaller machines will run out of memory
        #       and also take longer than typical trials
    ]
    invalid_ratios = [
        0.0,
        0.01,
        0.1,
        # TODO: Enable these additional values for the 'detailed' scenario
        # 0.5,
        # 0.9,
    ]
    ops = [operator.eq, operator.ne, operator.lt, operator.le, operator.ge, operator.gt]
    setup_params = itertools.product(
        rng_seeds,
        arr1_dtypes,
        arr2_dtypes,
        data_lengths,
        invalid_ratios,
        ops,
    )

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for (
        rng_seed,
        arr1_dtype,
        arr2_dtype,
        data_length,
        invalid_ratio,
        op,
    ) in setup_params:
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        arr1 = rand_array(rng, data_length, dtype=np.dtype(arr1_dtype), invalid_ratio=invalid_ratio)
        arr2 = rand_array(rng, data_length, dtype=np.dtype(arr2_dtype), invalid_ratio=invalid_ratio)

        # Sweep over other parameters that aren't required by the setup phase.
        other_params = [None]
        for _ in other_params:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)
            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):

                    start_time_ns = timestamper()

                    # invocation of actual actual function
                    op(arr1, arr2)

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "arr1_dtype": np.dtype(arr1_dtype),
                    "arr2_dtype": np.dtype(arr2_dtype),
                    "operation": op.__name__,
                    "data_length": data_length,
                    "invalid_ratio": invalid_ratio,
                    # Other parameters
                    # (None)
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_compare_ops_numpy(**kwargs):
    # TODO: Add additional dimensions:
    #  * number of threads
    #  * recycler on/off
    #  * different key multiplicity distributions (in the rand_fancyindex function)
    #  * different amounts of 'sortedness' of the fancy index (from the rand_fancyindex function)

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 0
    iters = 21  # This duration of this function is (usually) fairly short, so the performance is prone to random noise -- using more iterations helps

    # Parameters we'll sweep over for the benchmark.
    rng_seeds = [12345]
    arr1_dtypes = [
        np.int16,
        np.int32,
        np.float64,
    ]
    arr2_dtypes = [
        np.int16,
        np.int32,
        np.float64,
    ]
    data_lengths = [
        100,
        1_000,
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        # TODO: Add 100M, 1G and 2G -- these need to be optional since smaller machines will run out of memory
        #       and also take longer than typical trials
    ]
    ops = [operator.eq, operator.ne, operator.lt, operator.le, operator.ge, operator.gt]
    setup_params = itertools.product(
        rng_seeds,
        arr1_dtypes,
        arr2_dtypes,
        data_lengths,
        ops,
    )

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for (
        rng_seed,
        arr1_dtype,
        arr2_dtype,
        data_length,
        op,
    ) in setup_params:
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        arr1 = rand_array(rng, data_length, dtype=np.dtype(arr1_dtype))._np
        arr2 = rand_array(rng, data_length, dtype=np.dtype(arr2_dtype))._np
        # Sweep over other parameters that aren't required by the setup phase.
        other_params = [None]
        for _ in other_params:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):

                    start_time_ns = timestamper()

                    # invocation of actual actual function
                    op(arr1, arr2)

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "arr1_dtype": np.dtype(arr1_dtype),
                    "arr2_dtype": np.dtype(arr2_dtype),
                    "operation": op.__name__,
                    "data_length": data_length,
                    # Other parameters
                    # (None)
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def compare_mbget():
    return create_comparison_dataset(
        {
            "mbget": bench_mbget(),
            "mbget_numba": bench_mbget_numba(),
        }
    )


def compare_astype():
    return create_comparison_dataset(
        {
            "astype": bench_astype(),
            "astype_numpy": bench_astype_numpy(),
            "astype_numba": bench_astype_numba(),
        }
    )


def compare_bool_index():
    return create_comparison_dataset({"bool_index": bench_bool_index(), "bool_index_numpy": bench_bool_index_numpy()})


def compare_compare_ops():
    return create_comparison_dataset(
        {
            "compare_ops": bench_compare_ops(),
            "compare_ops_numpy": bench_compare_ops_numpy(),
        }
    )



import numpy as np

from .rt_fastarray import FastArray


class CompressedArray(FastArray):

    allowed_funcs = ["decompress", "view"]

    def __new__(cls, arr):
        if isinstance(arr, np.ndarray):
            arr.setflags(write=False)
        else:
            raise TypeError(f"Compressed array must be constructed from numpy array, not {type(arr)}")
        return arr.view(cls)

    def __init__(self, arr):
        pass

    def __getattribute__(self, attr):
        """
        Block all FastArray operations. See allowed_funcs class global.
        """
        if hasattr(FastArray, attr) and attr.startswith("_") is False and attr not in self.__class__.allowed_funcs:
            raise ValueError(f"{attr} cannot be accessed from compressed array.")
        return object.__getattribute__(self, attr)

    def decompress(self):
        pass

    def __repr__(self):
        return "test repr"

    def __str__(self):
        return "test str"

    def _build_string(self):
        # translate header information
        pass



import unittest

from riptable import *


class TimeWindow_Test(unittest.TestCase):
    def test_time_window(self):
        a = arange(100)
        r = rc.TimeWindow(uint64(a), int64(a), 0, 0)
        self.assertEqual(r[99], 99, msg=f"Wrong result produced for timewindow {r}")

        r = rc.TimeWindow(single(a), int64(a), 0, 0)
        self.assertEqual(r[99], 99.0, msg=f"Wrong result produced for timewindow {r}")

        r = rc.TimeWindow(int64(a), int64(a), 0, 0)
        self.assertEqual(r[99], 99, msg=f"Wrong result produced for timewindow {r}")

        r = rc.TimeWindow(int64(a), int64(a), 0, 3)
        self.assertEqual(r[99], 99 + 98 + 97 + 96, msg=f"Wrong result produced for timewindow {r}")


if __name__ == "__main__":
    tester = unittest.main()



import unittest

from riptable import *
from riptable.rt_datetime import NANOS_PER_DAY


def arr_eq(a, b):
    return arr_all(a == b)


def arr_all(a):
    return bool(np.all(a))


class RestoreSubclass_Test(unittest.TestCase):
    def test_tile(self):
        dtn = DateTimeNano.random(5)
        dtn2 = tile(dtn, 2)
        nptile = tile(dtn._np, 2)
        self.assertTrue(isinstance(dtn2, DateTimeNano))
        self.assertTrue(bool(np.all(dtn2._fa == nptile)))

        ts = TimeSpan(np.random.randint(0, NANOS_PER_DAY, 5, dtype=np.int64))
        ts2 = tile(ts, 2)
        nptile = tile(ts._np, 2)
        self.assertTrue(isinstance(dtn2, DateTimeNano))
        self.assertTrue(bool(np.all(ts2._fa == nptile)))

        d = Date(np.random.randint(15_000, 20_000, 5))
        d2 = tile(d, 2)
        nptile = tile(d._np, 2)
        self.assertTrue(isinstance(d2, Date))
        self.assertTrue(arr_eq(nptile, d2._fa))

        ds = DateSpan(np.random.randint(0, 365, 5))
        ds2 = tile(ds, 2)
        nptile = tile(ds._np, 2)
        self.assertTrue(isinstance(ds2, DateSpan))
        self.assertTrue(arr_eq(nptile, ds2._fa))

        c = Categorical(["a", "a", "b", "c", "a"])
        c2 = tile(c, 2)
        nptile = tile(c._np, 2)
        self.assertTrue(isinstance(c2, Categorical))
        self.assertTrue(arr_eq(nptile, c2._fa))

    def test_repeat(self):

        dtn = DateTimeNano.random(5)
        dtn2 = repeat(dtn, 2)
        nprep = repeat(dtn._np, 2)
        self.assertTrue(isinstance(dtn2, DateTimeNano))
        self.assertTrue(bool(np.all(dtn2._fa == nprep)))
        dtn2 = dtn.repeat(2)
        nprep = dtn._np.repeat(2)
        self.assertTrue(isinstance(dtn2, DateTimeNano))
        self.assertTrue(bool(np.all(dtn2._fa == nprep)))

        ts = TimeSpan(np.random.randint(0, NANOS_PER_DAY, 5, dtype=np.int64))
        ts2 = repeat(ts, 2)
        nprep = repeat(ts._np, 2)
        self.assertTrue(isinstance(dtn2, DateTimeNano))
        self.assertTrue(bool(np.all(ts2._fa == nprep)))
        ts2 = ts.repeat(2)
        nprep = ts._np.repeat(2)
        self.assertTrue(isinstance(dtn2, DateTimeNano))
        self.assertTrue(bool(np.all(ts2._fa == nprep)))

        d = Date(np.random.randint(15_000, 20_000, 5))
        d2 = repeat(d, 2)
        nprep = repeat(d._np, 2)
        self.assertTrue(isinstance(d2, Date))
        self.assertTrue(arr_eq(nprep, d2._fa))
        d2 = d.repeat(2)
        nprep = d._np.repeat(2)
        self.assertTrue(isinstance(d2, Date))
        self.assertTrue(arr_eq(nprep, d2._fa))

        ds = DateSpan(np.random.randint(0, 365, 5))
        ds2 = repeat(ds, 2)
        nprep = repeat(ds._np, 2)
        self.assertTrue(isinstance(ds2, DateSpan))
        self.assertTrue(arr_eq(nprep, ds2._fa))
        ds2 = ds.repeat(2)
        nprep = ds._np.repeat(2)
        self.assertTrue(isinstance(ds2, DateSpan))
        self.assertTrue(arr_eq(nprep, ds2._fa))

        c = Categorical(["a", "a", "b", "c", "a"])
        c2 = repeat(c, 2)
        nprep = repeat(c._np, 2)
        self.assertTrue(isinstance(c2, Categorical))
        self.assertTrue(arr_eq(nprep, c2._fa))
        c2 = c.repeat(2)
        nprep = c._np.repeat(2)
        self.assertTrue(isinstance(c2, Categorical))
        self.assertTrue(arr_eq(nprep, c2._fa))

    def test_gbkeys(self):
        length = 20
        ds = Dataset(
            {
                "CAT": Categorical(np.random.choice(["a", "b", "c"], length)),
                "DTN": DateTimeNano.random(length),
                "DATE": Date(np.random.randint(15000, 20000, length)),
                "TSPAN": TimeSpan(np.random.randint(0, 1_000_000_000 * 60 * 60 * 24, length, dtype=np.int64)),
                "DSPAN": DateSpan(np.random.randint(0, 365, length)),
            }
        )

        for k, v in ds.items():
            result = ds.gb(k).count()
            if k != "CAT":
                self.assertEqual(type(result[k]), type(v))
            self.assertTrue(arr_eq([k], result.label_get_names()))

    def test_mbget(self):
        length = 20
        ds = Dataset(
            {
                "CAT": Categorical(np.random.choice(["a", "b", "c"], length)),
                "DTN": DateTimeNano.random(length),
                "DATE": Date(np.random.randint(15000, 20000, length)),
                "TSPAN": TimeSpan(np.random.randint(0, 1_000_000_000 * 60 * 60 * 24, length, dtype=np.int64)),
                "DSPAN": DateSpan(np.random.randint(0, 365, length)),
            }
        )
        for k, v in ds.items():
            result = mbget(v, [1, 2, 3])
            self.assertEqual(type(result), type(v))

    def test_cat_grouping(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        ds = Dataset({"catcol": c, "data": FA([1, 1, 2, 3, 1])})
        ds2 = ds.drop_duplicates("data")

        c2 = ds2.catcol
        self.assertTrue(isinstance(c2, Categorical))
        self.assertTrue(arr_eq(["a", "b", "c"], c2))
        self.assertTrue(arr_eq([1, 2, 3], c2.grouping.ikey))
        self.assertTrue(c2.grouping.isdirty)
        self.assertTrue(arr_eq(c2[[0]], ["a"]))


if __name__ == "__main__":
    tester = unittest.main()



import unittest

import numpy as np

from riptable import FA, Categorical, FastArray, arange, cut, qcut


class Cut_Test(unittest.TestCase):
    """
    TODO: add more tests for different types
    also include string types
    """

    def test_cut(self):
        c = cut(arange(10), 3)
        self.assertTrue(sum(c._np - FA([1, 1, 1, 1, 2, 2, 2, 3, 3, 3])) == 0)

        c = cut(arange(10.0), 3)
        self.assertTrue(sum(c._np - FA([1, 1, 1, 1, 2, 2, 2, 3, 3, 3])) == 0)

        c = cut(arange(11), 3)
        self.assertTrue(sum(c._np - FA([1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3])) == 0)

        c = cut(FA([2, 4, 6, 8, 10]), FA([0, 2, 4, 6, 8, 10]))
        self.assertTrue(sum(c._np - FA([1, 2, 3, 4, 5])) == 0)

        c = cut(
            FA([2, 4, 6, 8, 10]),
            FA([0, 2, 4, 6, 8, 10]),
            labels=["a", "b", "c", "d", "e"],
        )
        self.assertTrue(sum(c._np - FA([1, 2, 3, 4, 5])) == 0)

        a = np.array([1, 7, 5, 4, 6, 3])
        l = FA([b"1.0->3.0", b"3.0->5.0", b"5.0->7.0"])

        c = cut(a, 3)
        self.assertIsInstance(c, Categorical)
        self.assertTrue(sum(c._np - FA([1, 3, 2, 2, 3, 1])) == 0)
        self.assertTrue((c.category_array == l).all())

        c = cut(a, 3, labels=True)
        self.assertIsInstance(c, Categorical)
        self.assertTrue(sum(c._np - FA([1, 3, 2, 2, 3, 1])) == 0)
        self.assertTrue((c.category_array == l).all())

        c = cut(a, 3, labels=None)
        self.assertIsInstance(c, Categorical)
        self.assertTrue(sum(c._np - FA([1, 3, 2, 2, 3, 1])) == 0)
        self.assertTrue((c.category_array == l).all())

        c = cut(a, 3, labels=False)
        self.assertIsInstance(c, FastArray)
        self.assertTrue(sum(c._np - FA([1, 3, 2, 2, 3, 1])) == 0)

        c, b = cut(a, 3, retbins=True)
        self.assertIsInstance(c, Categorical)
        self.assertIsInstance(b, np.ndarray)
        self.assertTrue(sum(c._np - FA([1, 3, 2, 2, 3, 1])) == 0)
        self.assertTrue((c.category_array == l).all())
        self.assertTrue(sum(b - FA([1.0, 3.0, 5.0, 7.0])) == 0)

        l = ["bad", "medium", "good"]
        c = cut(a, 3, labels=l)
        self.assertIsInstance(c, Categorical)
        self.assertTrue(sum(c._np - FA([1, 3, 2, 2, 3, 1])) == 0)
        self.assertTrue((c.category_array == l).all())

        # contiguous test
        x = arange(4).reshape(2, 2)
        knots = [-0.5, 0.5, 1.5, 2.5, 3.5]
        c = cut(x[:, 1], knots)
        l = FastArray([b"-0.5->0.5", b"0.5->1.5", b"1.5->2.5", b"2.5->3.5"])
        self.assertTrue((c.category_array == l).all())

        # inf upcast test
        x = np.array([0, 1, 10, 100, 5])
        knots = [-np.inf, 2, 11, 50, np.inf]
        c = cut(x, knots)
        self.assertTrue((c._fa == FA([1, 1, 2, 4, 2])).all())

        x = np.array([6, 3, 5, 2, 4, 1])
        c = cut(x, 2, filter=x < 5)
        self.assertTrue((c._fa == FA([0, 2, 0, 1, 2, 1])).all())

    def test_qcut(self):
        c = qcut(arange(10), 3)
        self.assertTrue(sum(c._np - FA([2, 2, 2, 2, 3, 3, 3, 4, 4, 4])) == 0)

        c = qcut(arange(11), 3)
        self.assertTrue(sum(c._np - FA([2, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4])) == 0)

        c = qcut(range(5), 3, labels=["good", "medium", "bad"])
        self.assertTrue(sum(c._np - FA([2, 2, 3, 4, 4])) == 0)

        c = qcut(arange(100.0), [0.05, 0.10, 0.25, 0.50, 0.75, 0.90, 0.95])
        self.assertTrue(c._np[0] == 1)
        self.assertTrue(c._np[5] == 2)
        self.assertTrue(c._np[94] == 7)
        self.assertTrue(c._np[95] == 1)

        c = qcut(arange(100.0), [0.00, 0.10, 0.25, 0.50, 0.75, 0.90, 0.95])
        self.assertTrue(c._np[0] == 2)
        self.assertTrue(c._np[5] == 2)
        self.assertTrue(c._np[94] == 7)
        self.assertTrue(c._np[95] == 1)

        c = qcut(range(5), 4)
        self.assertIsInstance(c, Categorical)
        self.assertTrue(sum(c._np - FA([2, 2, 3, 4, 5])) == 0)
        self.assertTrue((c.category_array == [b"Clipped", b"0.0->1.0", b"1.0->2.0", b"2.0->3.0", b"3.0->4.0"]).all())

        c = qcut(range(5), 4, labels=True)
        self.assertIsInstance(c, Categorical)
        self.assertTrue(sum(c._np - FA([2, 2, 3, 4, 5])) == 0)
        self.assertTrue((c.category_array == [b"Clipped", b"0.0->1.0", b"1.0->2.0", b"2.0->3.0", b"3.0->4.0"]).all())

        c = qcut(range(5), 4, labels=None)
        self.assertIsInstance(c, Categorical)
        self.assertTrue(sum(c._np - FA([2, 2, 3, 4, 5])) == 0)
        self.assertTrue((c.category_array == [b"Clipped", b"0.0->1.0", b"1.0->2.0", b"2.0->3.0", b"3.0->4.0"]).all())

        c = qcut(range(5), 3, labels=["good", "medium", "bad"])
        self.assertIsInstance(c, Categorical)
        self.assertTrue(sum(c._np - FA([2, 2, 3, 4, 4])) == 0)
        self.assertTrue(
            np.array(
                [
                    (h == t)
                    for (h, t) in zip(
                        c.expand_array.astype("U"),
                        ["good", "good", "medium", "bad", "bad"],
                    )
                ]
            ).all()
        )
        self.assertTrue((c.category_array == [[b"Clipped", b"good", b"medium", b"bad"]]).all())

        c = qcut(range(5), 4, labels=False)
        self.assertIsInstance(c, FastArray)
        self.assertTrue(sum(c._np - FA([2, 2, 3, 4, 5])) == 0)

        x = np.array([6, 3, 5, 2, 4, 1])
        c = qcut(x, 4, filter=x < 5)
        self.assertTrue((c._fa == FA([0, 4, 0, 3, 5, 2])).all())

    def test_cut_errors(self):
        with self.assertRaises(ValueError):
            c = cut(
                FA([2, 4, 6, 8, 10]),
                FA([0, 2, 4, 6, 8, 10]),
                labels=["a", "b", "c", "d", "e", "f"],
            )

    def test_qcut_drop_dups(self):
        c = qcut([1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 11, 12, 13], 4, duplicates="drop")
        self.assertIsInstance(c, Categorical)
        self.assertTrue(sum(c._np - FA([2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3])) == 0)
        self.assertTrue((c.category_array == [b"Clipped", b"1.0->2.0", b"2.0->13.0"]).all())


if __name__ == "__main__":
    tester = unittest.main()



from collections import Counter
from typing import List

import hypothesis
import numpy as np
import pytest
from hypothesis import HealthCheck, given
from hypothesis.extra.numpy import (
    arrays,
    byte_string_dtypes,
    datetime64_dtypes,
    integer_dtypes,
    timedelta64_dtypes,
    unicode_string_dtypes,
    unsigned_integer_dtypes,
)
from hypothesis.strategies import (
    booleans,
    composite,
    data,
    integers,
    lists,
    one_of,
    shared,
)

from riptable import Categorical, FastArray, hstack
from riptable.rt_enum import CategoryMode
from riptable.Utils.teamcity_helper import is_running_in_teamcity

from .strategies.categorical_strategy import CategoricalStrategy
from .strategies.helper_strategies import one_darray_shape_strategy

_MAX_SIZE = 1_000


def _get_category_to_count(categoricals) -> Counter:
    if not isinstance(categoricals, list):
        categorical = categoricals
        # create a new list since list constructor with categoricals will return the underlying representation as a list
        categoricals = list()
        categoricals.append(categorical)

    category_to_count = Counter()
    for categorical in categoricals:
        categories = categorical.categories()
        multiplicities = categorical.grouping.ncountgroup
        for category, multiplicity in zip(categories, multiplicities):
            category_to_count[category] += multiplicity

    return category_to_count


def _check_categorical(categorical: Categorical) -> (bool, str):
    valid: bool = True
    errors: List[str] = list()
    if not isinstance(categorical, Categorical):
        valid = False
        errors.append(f"Categorical {categorical} should be of type {type(Categorical)}")
    return valid, "\n".join(errors)


@composite
def one_of_categorical_values(draw):
    cat_values = integers(min_value=1, max_value=np.iinfo(np.int64).max)  # add - bytes(), characters(),
    return draw(one_of(cat_values))


@pytest.mark.skipif(
    reason="Categorical generator needs to be rewritten for better performance before re-enabling this test to run in TeamCity builds."
)
# 2020-09-09T21:02:22.3088485Z ================================== FAILURES ===================================
# 2020-09-09T21:02:22.3088746Z ________ test_categorical_ctor[CategoryMode.StringArray-integer_dtype] ________
# 2020-09-09T21:02:22.3088996Z
# 2020-09-09T21:02:22.3089415Z value_strategy = arrays(dtype=integer_dtypes(endianness='=', sizes=(64,)), shape=one_darray_shape_strategy(), elements=integers(min_value=1, max_value=9223372036854775807))
# 2020-09-09T21:02:22.3089925Z category_mode = <CategoryMode.StringArray: 1>
# 2020-09-09T21:02:22.3090026Z
# 2020-09-09T21:02:22.3090166Z >   ???
# 2020-09-09T21:02:22.3090416Z E   hypothesis.errors.FailedHealthCheck: Data generation is extremely slow: Only produced 9 valid examples in 1.13 seconds (0 invalid ones and 4 exceeded maximum size). Try decreasing size of the data you're generating (with e.g.max_size or max_leaves parameters).
# 2020-09-09T21:02:22.3091373Z E   See https://hypothesis.readthedocs.io/en/latest/healthchecks.html for more information about this. If you want to disable just this health check, add HealthCheck.too_slow to the suppress_health_check settings for this test.
@given(data())
@pytest.mark.parametrize(
    "value_strategy",
    [
        # Categorical values must be nonempty
        pytest.param(
            lists(one_of_categorical_values(), min_size=5, max_size=10),
            id="list",
        ),
        pytest.param(
            lists(one_of_categorical_values(), min_size=1, unique=True, max_size=10),
            id="unique_list",
        ),
        pytest.param(
            arrays(
                shape=one_darray_shape_strategy(max_shape_size=10),
                dtype=integer_dtypes(endianness="=", sizes=(64,)),
                elements=integers(min_value=1, max_value=np.iinfo(np.int64).max),
            ),
            id="integer_dtype",
        ),
        pytest.param(
            arrays(
                shape=one_darray_shape_strategy(),
                dtype=integer_dtypes(endianness="=", sizes=(64,)),
                elements=integers(min_value=1, max_value=np.iinfo(np.int64).max),
                fill=integers(min_value=0, max_value=np.iinfo(np.int64).max),
                unique=True,
            ),
            id="integer_dtype_unique",
            marks=[
                pytest.mark.skip,
                pytest.mark.xfail(
                    reason="Now throws a hypothesis.errors.InvalidArgument: Cannot fill unique array with non-NaN value 1"
                ),
            ],
        ),
    ],
)
@pytest.mark.parametrize("category_mode", [CategoryMode.StringArray, CategoryMode.Dictionary])
def test_categorical_ctor(value_strategy, category_mode, data):
    # cat is drawn from CategoricalStrategy
    ordered: bool = data.draw(booleans())
    cat: Categorical = data.draw(CategoricalStrategy(value_strategy, category_mode=category_mode, ordered=ordered))
    assert _check_categorical(cat)

    # Validate properties on constructing a Categorical from a Categorical's values and categories.
    values, categories = cat.expand_array, cat._categories
    # For Dictionary Categoricals, 'categories' should be the original Categorical's category_map.
    if category_mode == CategoryMode.Dictionary:
        categories = cat.category_mapping
    cat2 = Categorical(values, categories=categories, ordered=ordered)
    assert _check_categorical(cat2)

    # Validate properties on constructing a Categorical given a Categorical.
    cat3 = Categorical(cat2)
    assert _check_categorical(cat3)

    # Validate properties on constructing a Categorical using _from_categorical which is a fast path
    # that skips internal routine checks, sorting, or making values unique, but should be identical to
    # the original Categorical.
    from_categorical = cat._categories_wrap
    cat4 = Categorical(
        values,
        categories=categories,
        _from_categorical=from_categorical,
        ordered=ordered,
    )
    assert _check_categorical(cat4)

    # TODO: add equality checks for the Categoricals above since they should all be equivalent.


# TODO remove hypothesis suppress_health_check after investigating FailedHealthCheck for test_categorical_property.test_hstack[CategoryMode_StringArray-unsigned_integer_dtype]
# E   hypothesis.errors.FailedHealthCheck: Data generation is extremely slow: Only produced 7 valid examples in 1.06 seconds (0 invalid ones and 5 exceeded maximum size). Try decreasing size of the data you're generating (with e.g.max_size or max_leaves parameters).
# As is, the unsigned_integer_dtype case uses min and max values for data generation.
@pytest.mark.skip(
    reason="Categorical generator needs to be rewritten for better performance before re-enabling this test to run in TeamCity builds."
)
@hypothesis.settings(suppress_health_check=[HealthCheck.too_slow])
@given(data())
@pytest.mark.parametrize(
    "datatype, elements",
    [
        pytest.param(
            integer_dtypes(endianness="=", sizes=(64,)),
            integers(min_value=1, max_value=np.iinfo(np.int64).max),
            id="integer_dtype",
        ),
        pytest.param(
            unsigned_integer_dtypes(endianness="=", sizes=(64,)),
            integers(min_value=1, max_value=np.iinfo(np.int64).max),
            id="unsigned_integer_dtype",
        ),
        pytest.param(byte_string_dtypes(endianness="="), None, id="byte_string_dtype"),
        pytest.param(
            datetime64_dtypes(endianness="="),
            None,
            id="datetime64_dtype",
            marks=[
                pytest.mark.xfail(reason="RIP-375 - Categorical unsupported dtypes"),
                pytest.mark.skip,
            ],
        ),
        pytest.param(
            timedelta64_dtypes(endianness="="),
            None,
            id="timedelta64_dtype",
            marks=[
                pytest.mark.xfail(reason="RIP-375 - Categorical unsupported dtypes"),
                pytest.mark.skip,
            ],
        ),
        pytest.param(
            unicode_string_dtypes(endianness="="),
            None,
            id="unicode_string_dtype",
            marks=[
                pytest.mark.xfail(reason="RIP-375 - Categorical unsupported dtypes"),
                pytest.mark.skip,
            ],
        ),
    ],
)
@pytest.mark.parametrize("category_mode", [CategoryMode.StringArray])
def test_hstack(datatype, elements, category_mode, data):
    shape = one_darray_shape_strategy()
    dtype = shared(datatype)
    msg = f"Using dtype {dtype}\nUsing elements {elements}\n"

    # Increasing the maximum number of runs by a 10x magnitude will result in FailedHealthCheck errors with slow data generation.
    max = data.draw(integers(min_value=1, max_value=5))
    categoricals: List[Categorical] = list()
    for i in range(max):
        value_strategy = arrays(dtype, shape, elements=elements)
        with_categories: bool = data.draw(booleans())
        categoricals.append(
            data.draw(
                CategoricalStrategy(
                    value_strategy,
                    with_categories=with_categories,
                    category_mode=category_mode,
                )
            )
        )

    # Test #1: Length of hstacked categoricals should be the sum of the aggregate categoricals.
    output = hstack(tuple(categoricals))
    assert isinstance(output, Categorical)

    assert len(output) == sum(map(len, categoricals)), (
        f"Length of hstacked categoricals should be the sum of the aggregate categoricals\n"
        + msg
        + f"actual:\n{output}\nexpected:\n{categoricals}"
    )

    # Test #2: The hstacked categories should be equivalent to the set of aggregate categories.
    expected_counts = _get_category_to_count(categoricals)
    actual_counts = _get_category_to_count(output)

    assert not set(actual_counts.elements()).symmetric_difference(set(expected_counts.elements())), (
        f"The hstacked categories should be equivalent to the set of aggregate categories\n"
        + msg
        + f"actual {set(actual_counts.elements())}\nexpected {set(expected_counts.elements())}"
    )

    # Test #3: The hstacked multiplicity of categories should be equivalent to the multiplicity of aggregate categories.
    # Test (2) is a subset of this equality check, but remains for clarity reasons when investigating failures.
    assert expected_counts == actual_counts, (
        f"The hstacked multiplicity of categories should be equivalent to the multiplicity of aggregate categories\n"
        + msg
        + f"actual {actual_counts}\nexpected {expected_counts}"
    )


@pytest.mark.xfail(reason="RIP-375 - Categorical unsupported dtypes")
@pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
@pytest.mark.parametrize(
    "data",
    [
        # ValueError: BuildArrayInfo array has bad dtype of 21
        FastArray(["1970"], dtype="datetime64[Y]"),
        # ValueError: BuildArrayInfo array has bad dtype of 22
        FastArray([0], dtype="timedelta64[Y]"),
    ],
)
def test_falsifying_categorical_ctor(data):
    Categorical(data)


@pytest.mark.skipif(True, reason="RIP-452: Mutikey Categorical isin is consistent with its single key isin alternative")
def test_multikey_categorical_isin():
    # See Python/core/riptable/tests/test_categorical.py test_multikey_categorical_isin as an example
    pass
